/*
 * Micron Mt9M112 camera driver.
 *
 * Copyright (C) 2005-2007 Atmel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
//#define DEBUG

#include <linux/clk.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/device.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <linux/i2c.h>

#include <linux/err.h>
#include <asm/gpio.h>



#include <mach/at32ap700x.h>
#include <mach/board.h>
//#include <mach/hmatrix.h>
//#include <mach/pm.h>

#include <mach/portmux.h>
#include <media/atmel-isi.h>


#define select_peripheral(port, pin_mask, periph, flags)	\
	at32_select_periph(GPIO_##port##_BASE, pin_mask,	\
			   GPIO_##periph, flags)


#define PORTA (void*)0xFFE02800
#define PORTB (void*)0xFFE02C00
#define PORTC (void*)0xFFE03000
#define PORTD (void*)0xFFE03400
#define PORTE (void*)0xFFE03800

#define PIO_PER 0x0000//PIO Enable
#define PIO_OER 0x0010//Output Enable
#define PIO_ODR 0x0014//Output disdable

#define PIO_SODR 0x0030//Set output
#define PIO_CODR 0x0034//clear output
#define PIO_PDSR 0x003C

#define PIO_PUDR 0x0060

/* camera standby pin */

/* camera reset pin */
#define CAM_RESET (1<<18)

/*! Maximum number of pixels in a row */
#define tcm8230_MAX_WIDTH	640
/*! Maximum number of rows */
#define tcm8230_MAX_HEIGHT	480

#define I2C_DRIVERID_tcm8230	93

/*! Clock for image sensor CLKIN signal */
static char mclk_name[32] = "gclk0";
/*! Parent clock of gclk0 
 *  Either osc0 or pll0
 *  We use osc0 with 20MHz quarz.
 */
static char mclk_parent_name[32] = "osc0";

static struct clk *mclk;
static struct clk *mclk_parent;
module_param_string(mclk, mclk_name, sizeof(mclk_name), 0644);
MODULE_PARM_DESC(mclk, "Name of the clock used as camera clock input");

module_param_string(mclk_parent, mclk_parent_name,
		    sizeof(mclk_parent_name), 0644);
MODULE_PARM_DESC(mclk, "Name of mclk parent clock");




//struct tm13m3 {
struct tcm8230 {
	struct mutex		mutex;
	u16			current_page;
	u32			current_format;
	u16			pll_avr_ctrl;
	struct clk		*mclk;
	struct i2c_client	client;
	struct atmel_isi_camera	cam;
};


#define to_tcm8230(cam) container_of(cam, struct tcm8230, cam)
//**************************************************************************************************************************************
#define i2c_sda (1<<7)
#define i2c_scl (1<<6)
#define i2c_mask (1<<6)|(1<<7)
/* Chip ID stored in CHIP_VERSION register */
#define MT9M112_CHIP_ID 0x70
/* I2C address of camera module */
#define I2C_tcm8230	0x3C

#define I2C_DELAY 1000
#define I2C_DELAY_ON 200

static int i2c_start(void)
{
	//PA6 SDA
	//PA7 SCL
	__raw_writel(i2c_mask, PORTA + PIO_PER);
	__raw_writel(i2c_mask, PORTA + PIO_OER);
	__raw_writel(i2c_mask, PORTA + PIO_SODR);

	__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Data off
	__raw_writel(i2c_scl, PORTA + PIO_SODR);//pioa->sodr = I2C_SCL_MASK;//Colck On
	udelay(I2C_DELAY_ON);
	//__raw_writel(i2c_sda, PORTA + PIO_SODR);//pioa->sodr = I2C_SDA_MASK;//Data on
	__raw_writel(i2c_sda, PORTA + PIO_ODR);//pioa->odr = I2C_SDA_MASK;//Output disdable Register ODR

	//__raw_writel(i2c_mask, PORTA + PIO_CODR);
	return 0;
}

static void i2c_clock_pulse(void)
{

	__raw_writel(i2c_sda, PORTA + PIO_OER);//pioa->oer=I2C_SDA_MASK;//Output enable DATA
	__raw_writel(i2c_scl, PORTA + PIO_SODR);//pioa->sodr = I2C_SCL_MASK;// Clock On
	udelay(I2C_DELAY);//tHIGH
	
	__raw_writel(i2c_scl, PORTA + PIO_CODR);//pioa->codr = I2C_SCL_MASK;// Clock Off
	__raw_writel(i2c_sda, PORTA + PIO_ODR);//pioa->odr = I2C_SDA_MASK;//Output disdable Data
	udelay(I2C_DELAY);
}
static int i2c_ack_inp(void)
{
	int i=0;
	
	//__raw_writel(i2c_sda, PORTA + PIO_ODR);//pioa->odr = I2C_SDA_MASK;//Output disdable Register ODR
	//udelay(I2C_DELAY);
	__raw_writel(i2c_scl, PORTA + PIO_SODR);//pioa->sodr = I2C_SCL_MASK;//Colck On
	//udelay(I2C_DELAY);
	while(__raw_readl(PORTA + PIO_PDSR) & i2c_sda)
	{
		i++;
		if(i>100000)
		{
			__raw_writel(i2c_scl, PORTA + PIO_CODR);//pioa->codr = I2C_SCL_MASK;//Clock Off
			return 1;
		}
	}
	udelay(I2C_DELAY);
	__raw_writel(i2c_scl, PORTA + PIO_CODR);//pioa->codr = I2C_SCL_MASK;//Clock Off
	//__raw_writel(i2c_sda, PORTA + PIO_OER);//pioa->oer=I2C_SDA_MASK;//Output enable DATA
	
	return 0;
}
static int i2c_send(unsigned char device_adresse,unsigned char device_reg,unsigned char data)
{
	unsigned char i=0,dd=0x40;
	char ret=0;
	
	//Start Ubertragung
	__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Data Off
	__raw_writel(i2c_sda, PORTA + PIO_OER);//pioa->oer=I2C_SDA_MASK;//Output enable DATA

	udelay(I2C_DELAY_ON);
	__raw_writel(i2c_scl, PORTA + PIO_CODR);//pioa->codr = I2C_SCL_MASK;//Clock Off
	udelay(I2C_DELAY);
	//******************************************************************************
	__raw_writel(i2c_sda, PORTA + PIO_ODR);//pioa->odr = I2C_SDA_MASK;//Output disdable Register ODR

	//Send addresse
	dd=0x40;
	for (i=0;i<7;i++)
	{
		if(dd & device_adresse)
			__raw_writel(i2c_sda, PORTA + PIO_SODR);//pioa->sodr = I2C_SDA_MASK;// Set Output Data Register
		else
			__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;// Set Output Data Register

		i2c_clock_pulse();
		dd=(dd>>1);
	}
	//Bit Read/Write Read=>1 Write=>0
	__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Write SDA=1
	i2c_clock_pulse();

	//ACK
	if(i2c_ack_inp())
	{
		ret =1;
		goto ende;
	}
	//send Register auf dem Chip
	dd=0x80;
	for (i=0;i<8;i++)
	{
		if(dd & device_reg)
			__raw_writel(i2c_sda, PORTA + PIO_SODR);//pioa->sodr = I2C_SDA_MASK;//Data On
		else
			__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Data Off
		i2c_clock_pulse();
		dd=(dd>>1);
	}
	
	//ACK
	if(i2c_ack_inp())
	{
		ret =2;
		goto ende;
	}

	//Daten senden
	dd=0x80;
	for (i=0;i<8;i++)
	{
		if(dd & data)
			__raw_writel(i2c_sda, PORTA + PIO_SODR);//pioa->sodr = I2C_SDA_MASK;//Data On
		else
			__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Data Off
		i2c_clock_pulse();
		dd=(dd>>1);
	}
	
	//ACK
	if(i2c_ack_inp())
	{
		ret =3;
		goto ende;
	}

ende:
	
	__raw_writel(i2c_sda, PORTA + PIO_CODR);//pioa->codr = I2C_SDA_MASK;//Data off
	__raw_writel(i2c_sda, PORTA + PIO_OER);//pioa->oer=I2C_SDA_MASK;//Output enable DATA

	__raw_writel(i2c_scl, PORTA + PIO_SODR);//pioa->sodr = I2C_SCL_MASK;//Colck On
	udelay(I2C_DELAY_ON);
	//__raw_writel(i2c_sda, PORTA + PIO_SODR);//pioa->sodr = I2C_SDA_MASK;//Data on
	__raw_writel(i2c_sda, PORTA + PIO_ODR);//pioa->odr = I2C_SDA_MASK;//Output disdable Register ODR
	udelay(I2C_DELAY);
	return ret;
}




static int tcm8230_get_format(struct atmel_isi_camera *cam,
			   struct atmel_isi_format *fmt)
{
	struct tcm8230 *is = to_tcm8230(cam);
	int ret = 0;

	fmt->pix.colorspace = V4L2_COLORSPACE_SMPTE170M;
	fmt->input_format = ATMEL_ISI_PIXFMT_RGB16;//ATMEL_ISI_PIXFMT_CbYCrY;//
	fmt->input_format = is->current_format;

	fmt->pix.width = 640;
	fmt->pix.height = 480;

	return ret;
}

static int tcm8230_set_format(struct atmel_isi_camera *cam,
			   struct atmel_isi_format *fmt)
{
	struct tcm8230 *is = to_tcm8230(cam);
	int ret = 0;

	fmt->pix.colorspace = V4L2_COLORSPACE_SMPTE170M;

	fmt->input_format = ATMEL_ISI_PIXFMT_RGB16;//ATMEL_ISI_PIXFMT_CrYCbY;
	is->current_format = ATMEL_ISI_PIXFMT_RGB16;//ATMEL_ISI_PIXFMT_CrYCbY;

	/* adjust picture width and height */
	if (fmt->pix.width > tcm8230_MAX_WIDTH)
		fmt->pix.width = tcm8230_MAX_WIDTH;
	if (fmt->pix.height > tcm8230_MAX_HEIGHT)
		fmt->pix.height = tcm8230_MAX_HEIGHT;

	/* FIXME Set context output width needed ??*/

	pr_debug("%s: set_format %ux%u\n", cam->name,
		 fmt->pix.width, fmt->pix.height);
	return ret;
}


static void tcm8230_reset_hardware(struct tcm8230 *is)
{
	__raw_writel(CAM_RESET, PORTC + PIO_CODR);
	//FIXME : set correct reset interval usleep();
	udelay(1000);
	__raw_writel(CAM_RESET, PORTC + PIO_SODR);
	udelay(10000);
	udelay(10000);
}

static int tcm8230_start_capture(struct atmel_isi_camera *cam)
{
	struct tcm8230 *is = to_tcm8230(cam);
	int ret = 0;

	return ret;
}

static int tcm8230_stop_capture(struct atmel_isi_camera *cam)
{
	struct tcm8230 *is = to_tcm8230(cam);
	int ret = 0;


	return ret;
}

static int tcm8230_init_hardware(struct tcm8230 *is)
{
	int chip_id;
	int ret;
	int i=0;
	tcm8230_reset_hardware(is);
	udelay(10000);
	udelay(10000);
	udelay(10000);
	udelay(10000);
	udelay(10000);

	/* set register page to reset value*/
	is->current_page = 0;
	//is->current_format = ATMEL_ISI_PIXFMT_CbYCrY;
	is->current_format = ATMEL_ISI_PIXFMT_RGB16;

	i2c_start();
	udelay(10000);
	udelay(10000);
	udelay(10000);
	udelay(10000);
	udelay(10000);
	udelay(10000);
	
	for (i=0;i<128;i++)
	{
		if(i2c_send(i,0x0,0x0)==0)
			printk(KERN_INFO "i2c Device fount: 0x%x\n",i);
	
	}

	
	udelay(10000);
	
	ret=i2c_send(0x3C,0x3,0x2);
	if(ret)
	{
		printk(KERN_INFO "TCM8230: reg 0x3 Kamera return Error: %i\n",ret);
		return 	-1;
	}
	
	/*ret=i2c_send(0x3C,0x4,0xcf);
	if(ret)
	{
		printk(KERN_INFO "Kamera return Error: %i\n",ret);
		return 	-1;
	}*/
	ret=i2c_send(0x3C,0x1e,0x40);//0x40 HSYNCSEL Off....0x48 ON
	if(ret)
	{
		printk(KERN_INFO "TCM8230: reg 0x1E Kamera return Error: %i\n",ret);
		return 	-1;
	}


	printk(KERN_INFO "Kamera init\n");
	return 0;
}



static int __init tcm8230_init(void)
{
	        /*
         * Set up the master clock, if available. If clk_get() fails,
         * this hopefully means that the board generates a suitable
         * master clock some other way, which is fine by us.
         *
         * We need to do this before probing the i2c bus, as the
         * camera won't ack any messages when it doesn't have a clock.
         */
	struct tcm8230 *is;
	int ret;


	//select_peripheral(PIOA, (1<<30), PERIPH_A, 0);

        mclk_parent = clk_get(NULL, mclk_parent_name);
        if (!IS_ERR(mclk_parent))
                clk_enable(mclk_parent);
        else {
                mclk_parent = NULL;
		pr_debug("tcm8230: No parent clock available\n");
	}

        mclk = clk_get(NULL, mclk_name);
        if (!IS_ERR(mclk)) {
                if (mclk_parent)
                        clk_set_parent(mclk, mclk_parent);

                clk_set_rate(mclk, 20000000);
                clk_enable(mclk);
        } else {
                mclk = NULL;
		pr_debug("tcm8230: No clock set\n");
        }

	
	/* Reset sequence */
	__raw_writel(CAM_RESET, PORTC + PIO_PER);
        __raw_writel(CAM_RESET, PORTC + PIO_OER);
	//__raw_writel(CAM_RESET, PORTC + PIO_CODR);
	//udelay(1000);
	__raw_writel(CAM_RESET, PORTC + PIO_SODR);

	
	is = kzalloc(sizeof(struct tcm8230), GFP_KERNEL);
	if (!is)
		return -ENOMEM;

	
	is->cam.name = "tcm8230";
	is->cam.hsync_act_low = 0;
	is->cam.vsync_act_low = 0;
	is->cam.pclk_act_falling = 0;
	
	/* no SAV/EAV sync -> HSYNC and VSYNC used */
	/*is->cam.has_emb_sync = 0;*/
	is->cam.has_emb_sync = 0;

	is->cam.get_format = tcm8230_get_format;
	is->cam.set_format = tcm8230_set_format;
	is->cam.start_capture = tcm8230_start_capture;
	is->cam.stop_capture = tcm8230_stop_capture;

	mutex_init(&is->mutex);

	is->mclk = clk_get(NULL, mclk_name);
	if (IS_ERR(is->mclk)) {
		ret = PTR_ERR(is->mclk);
		goto err_clk;
	}
	clk_enable(is->mclk);

	
	ret = tcm8230_init_hardware(is);
	if (ret)
		goto err_init_hw;

	/* We're up and running. Notify the ISI driver */
	ret = atmel_isi_register_camera(&is->cam);
	if (ret)
		goto err_register;

	printk(KERN_INFO "tcm8230 Image Sensor at %s\n", is->cam.name);

	return ret;

err_register:
err_init_hw:
	clk_disable(is->mclk);
	clk_put(is->mclk);

err_clk:
	kfree(is);
	return ret;

}
module_init(tcm8230_init);

static void __exit tcm8230_exit(void)
{
	if (mclk) {
		clk_disable(mclk);
		clk_put(mclk);
	}
	if (mclk_parent) {
		clk_disable(mclk_parent);
		clk_put(mclk_parent);
	}
	
}
module_exit(tcm8230_exit);

MODULE_DESCRIPTION("Atmel Image Sensor Interface Driver");
MODULE_AUTHOR("Heinrich Landgraf");
MODULE_LICENSE("GPL");
